"""DAG related utilities."""
from __future__ import annotations

# std
import time
from typing import Any, Optional

# external
from redis import Redis
import rq
from rq.queue import Queue
from rq.worker import Worker

# module
from ._constants import DAG_INDEX, DAG_OPERATIONS, DAG_STATUS, hash_t, join, OPERATIONS
from ._context import get_storage
from ._graph import Artefact, get_op_options, Operation, resolve_link
from ._logging import logger
from ._run import run_op, RunStatus
from ._short_hash import shorten_hash


def __set_as_hashes(db: Redis[bytes], key1: str, key2: str) -> set[hash_t]:
    mem = db.sinter(key1, key2)
    out = set()
    for k in mem:
        if isinstance(k, bytes):
            out.add(hash_t(k.decode()))
        elif isinstance(k, str):
            out.add(hash_t(k))
        else:
            out.add(hash_t(str(k)))
    return out


def _dag_dependents(db: Redis[bytes], dag_of: hash_t, op_from: hash_t) -> set[hash_t]:
    """Get dependents of an op within a given DAG."""
    return __set_as_hashes(
        db, join(DAG_OPERATIONS, dag_of), join(OPERATIONS, op_from, "children")
    )


def delete_all_dags(db: Redis[bytes]) -> None:
    """Delete all currently stored DAGs."""
    for dag in db.smembers(DAG_INDEX):
        db.delete(join(DAG_OPERATIONS, dag.decode()))  # type:ignore
        db.delete(join(DAG_STATUS, dag.decode()))  # type:ignore
    # Remove old index
    db.delete(DAG_INDEX)


def ancestors(
    db: Redis[bytes], *addresses: hash_t, include_subdags: bool = False
) -> set[hash_t]:
    """Get all ancestors of a given hash."""
    queue = list(addresses)
    out = set()

    while len(queue) > 0:
        curr = queue.pop()
        parents = db.smembers(join(OPERATIONS, curr, "parents"))

        if include_subdags:
            # special case for subdag operations
            parents = parents.union(
                db.smembers(join(OPERATIONS, curr, "parents.subdag"))
            )

        for el in parents:
            if el == b"root":
                continue

            h = hash_t(el.decode())
            out.add(h)
            if h not in queue:
                queue.append(h)
    return out


def descendants(db: Redis[bytes], *addresses: hash_t) -> set[hash_t]:
    """Get all descendants of a given hash."""
    queue = list(addresses)
    out = set()

    while len(queue) > 0:
        curr = queue.pop()
        for el in db.smembers(join(OPERATIONS, curr, "children")):
            h = hash_t(el.decode())
            out.add(h)
            if h not in queue:
                queue.append(h)
    return out


def get_nearest_operation(
    db: Redis[bytes], address: hash_t, subdag: Optional[str] = None
) -> Optional[Operation]:
    """Return the operation at address or the operation generating address."""
    root = "root"
    art = None
    try:
        node = Operation.grab(db, address)
        return node
    except RuntimeError:
        # one possibility is that address is an artefact...
        try:
            art = Artefact[Any].grab(db, address)
        except RuntimeError:
            raise RuntimeError(
                f"address {address} neither a valid operation nor a valid artefact."
            )

    if art.parent == root:
        # We have basically just a single artefact as the network...
        return None
    else:
        node = Operation.grab(db, art.parent)
        return node


def build_dag(
    db: Redis[bytes], address: hash_t, subdag: Optional[str] = None
) -> None:  # noqa:C901
    """Setup DAG required to compute the result at a specific address."""
    node = get_nearest_operation(db, address)
    if node is None:
        return

    # Ok, so now we finally know we have a node, and we want to extract the whole DAG
    # from it.
    ancs = ancestors(db, node.hash)
    logger.debug(f"{node.hash[:6]} has {len(ancs)} ancestors")

    if subdag is None:
        dag_of = address
    else:
        dag_of = hash_t(f"{subdag}/{address}")

    table = join(DAG_STATUS, dag_of)
    ops = join(DAG_OPERATIONS, dag_of)

    # Initialize the dependencies count for each DAG operation.
    pipe = db.pipeline(transaction=True)
    pipe.delete(table)  # get rid of previous status data
    for address in ancs.union([node.hash]):
        ndepen = db.scard(join(OPERATIONS, address, "parents"))
        pipe.hset(table, address, ndepen)
        pipe.sadd(ops, address)

    pipe.sadd(DAG_INDEX, dag_of)
    pipe.execute()


def enqueue_dependents(
    dag_of: hash_t,
    current: hash_t,
) -> None:
    """Enqueue dependents."""
    job = rq.get_current_job()
    db: Redis[bytes] = job.connection
    depen = _dag_dependents(db, dag_of, current)
    logger.info(f"has {len(depen)} dependents")

    dagtable = join(DAG_STATUS, dag_of)
    for dependent in depen:
        # First, atomically update dependencies status
        ndepen = db.hincrby(dagtable, dependent, -1)

        if ndepen == 0:
            # Operation is ready to be executed.
            options = get_op_options(db, dependent)
            queue = Queue(options.queue, connection=db, **options.queue_args)

            logger.info(f"-> {shorten_hash(dependent)}")
            queue.enqueue_call(
                "funsies._dag.task",
                args=(dag_of, dependent),
                kwargs=options.task_args,
                **options.job_args,
            )

    # We may want to execute a subdag dependent
    components = dag_of.split("/")
    if len(components) > 1:
        evaluating = components[-1]
        from_op = components[-2]
        in_parent_dag = components[:-2]
        if current == evaluating:
            logger.info("done evaluating subdag")
            logger.info(f"enqueuing dependents of {shorten_hash(hash_t(from_op))}")
            logger.info(
                "within dag of"
                + f' {"/".join([shorten_hash(hash_t(el)) for el in in_parent_dag])}'
            )
            enqueue_dependents(hash_t("/".join(in_parent_dag)), hash_t(from_op))


def acquire_task(db: Redis[bytes], op_hash: hash_t, worker_name: Optional[str]) -> bool:
    """Check if someone else is currently executing this job."""
    if worker_name is None:
        # running in non-distributed mode
        return True

    owner_key = join(OPERATIONS, op_hash, "owner")
    response = db.setnx(owner_key, worker_name)
    if response:
        return True
    else:
        key = db.get(owner_key)
        if key is None:
            # the other worker just just drop off like right now
            # to avoid a race condition, will wait till later.
            logger.info("issue acquiring lock, will try again later")
            return False

        holder = key.decode()
        logger.info(f"job currently held by {holder}")
        if holder == worker_name:
            logger.error("other worker is myself! HOW!?")
            return True

        # grab other holder
        worker = Worker.find_by_key(
            Worker.redis_worker_namespace_prefix + holder, connection=db
        )
        if worker is None:
            # holder is gooooonnnneee. let's take over.
            db.set(owner_key, worker_name)
            logger.warning("other worker is gone, I'm taking over")
            return True
        else:
            if worker.state == "busy":
                # get other worker's job
                ojob = worker.get_current_job()
                if op_hash in ojob.description:
                    logger.info("will try again later")
                    return False
                else:
                    db.set(owner_key, worker_name)
                    logger.error("other worker has moved on, I'm taking over")
                    return True
            else:
                db.set(owner_key, worker_name)
                logger.error("other worker is not working, I'm taking over")
                return True


def task(
    dag_of: hash_t,
    current: hash_t,
    *,
    evaluate: bool = True,
) -> RunStatus:
    """Worker evaluation of a given step in a DAG."""
    # load database
    job = rq.get_current_job()
    db: Redis[bytes] = job.connection
    worker_name: Optional[str] = job.worker_name
    logger.debug(f"attempting {current} on {worker_name}.")

    # TODO: Fix
    store = get_storage(None)

    # Run job
    with logger.contextualize(op=shorten_hash(current)):
        # Start by checking if another worker is currently executing this operation
        acquired = acquire_task(db, current, worker_name)
        if not acquired:
            # Do job later
            time.sleep(0.5)  # delay so as to not hit the db too often
            options = get_op_options(db, current)
            queue = Queue(name=options.queue, connection=db, **options.queue_args)
            queue.enqueue_call(
                "funsies._dag.task",
                args=(dag_of, current),
                kwargs=options.task_args,
                at_front=False,
                **options.job_args,
            )
            stat = RunStatus.delayed
        else:
            # Run operation
            op = Operation.grab(db, current)
            stat = run_op(db, store, op, evaluate=evaluate)

            if stat == RunStatus.subdag_ready:
                # We have created a subdag
                for value in op.out.values():
                    ln = resolve_link(db, value)
                    art = Artefact[Any].grab(db, ln)
                    logger.info(f"starting subdag -> {shorten_hash(art.parent)}")
                    start_dag_execution(db, art.parent, subdag=f"{dag_of}/{current}")

            if stat > 0:
                # Success! Let's (possibly) enqueue dependents.
                enqueue_dependents(dag_of, current)

    # reset lock
    db.delete(join(OPERATIONS, current, "owner"))
    return stat


def start_dag_execution(
    db: Redis[bytes], data_output: hash_t, subdag: Optional[str] = None
) -> None:
    """Execute a DAG to obtain a given output using an RQ queue."""
    # make dag
    build_dag(db, data_output, subdag)

    if subdag is not None:
        dag_of = hash_t(f"{subdag}/{data_output}")
    else:
        dag_of = data_output

    # enqueue everything starting from root
    for element in _dag_dependents(db, dag_of, hash_t("root")):
        options = get_op_options(db, element)
        queue = Queue(name=options.queue, connection=db, **options.queue_args)
        queue.enqueue_call(
            "funsies._dag.task",
            args=(dag_of, element),
            kwargs=options.task_args,
            **options.job_args,
        )



"""User-friendly interfaces to funsies functionality."""
from __future__ import annotations

# std
from typing import Any, Callable, Iterable, Mapping, Optional, Sequence, TypeVar, Union

# external
from redis import Redis

# module
from ._constants import _AnyPath, _Data, Encoding
from ._context import Connection, get_connection, get_options
from ._graph import Artefact, constant_artefact, make_op
from ._infer import output_types
from ._pyfunc import python_funsie
from .config import Options, StorageEngine

# Types
_Target = Union[Artefact, _Data]
_INP_FILES = Optional[Mapping[_AnyPath, _Target]]
_OUT_FILES = Optional[Iterable[_AnyPath]]
T = TypeVar("T")


def _artefact(
    db: Redis[bytes], store: StorageEngine, data: Union[T, Artefact[T]]
) -> Artefact[T]:
    if isinstance(data, Artefact):
        return data
    else:
        return constant_artefact(db, store, data)


# Yay overloads! we all wish there were an easier way of doing this but here we are...
Tin1 = TypeVar("Tin1", bound=_Data)
Tin2 = TypeVar("Tin2", bound=_Data)
Tin3 = TypeVar("Tin3", bound=_Data)
Tin4 = TypeVar("Tin4", bound=_Data)
Tout1 = TypeVar("Tout1", bound=_Data)
_inp = Union[Tin1, Artefact[Tin1]]


# --------------------------------------------------------------------------------
# Data transformers
def py(  # noqa:C901
    fun: Callable[..., Any],
    *inp: _Target,
    out: Optional[Sequence[Encoding]] = None,
    name: Optional[str] = None,
    strict: bool = True,
    opt: Optional[Options] = None,
    connection: Connection = None,
) -> Union[Artefact, tuple[Artefact, ...]]:
    """Add a python function to the workflow.

    `py(fun, *inp)` puts a python function `fun` on the workflow and returns
    its output artefact.

    As many arguments will be passed to `fun()` as there are input
    `types.Artefact` instances in `*inp` and `fun()` should return as many
    outputs as there are data types in `out=`. By default, `out=` will be
    inferred from annotations.

    If `strict=False`, the function is taken to do it's own error handling and
    arguments will be of type `errors.Result[T]` instead of `T`. See
    `utils.match_results()` for a convenient way to process these values.

    Python function hashes are generated based on their names (as given by
    `fun.__qualname__`) and functions are distributed to workers using
    `cloudpickle`. This is important because it means that:

    - Workers must have access to the function if it is imported, and must
        have access to any imported libraries.

    - Changing a function without modifiying its name (or modifying the
        `name=` argument) will not recompute the graph.

    It is the therefore the caller's responsibility to `reset()` one of the
    return value of `py()` if the function is modified to ensure re-excution
    of its dependents.

    Args:
        fun: Python function that operates on input artefacts and produces a
            single output artefact.
        *inp: Input artefacts.
        out: List of Encoding, one for each output of fun. These are the kind
            of serialization-deserialization used for the output variables. If
            None, `out=` is inferred using the type hint of `fun()`. It is
            `types.Encoding.blob` for all `bytes` outputs and
            `types.Encoding.json` for anything else.
        name: Override the name of `fun()` used in hash generation.
        strict: If `False`, error handling will be deferred to `fun()` by
            passing it argument of type `errors.Result[bytes]` instead of
            `bytes`.
        connection: An explicit Redis connection. Not required if called
            within a `Fun()` context.
        opt: An `types.Options` instance generated from `options()`. Not
            required if called within a `Fun()` context.

    Returns:
        A single `types.Artefact` instance if `out=` contains only one element
        or a tuple of `types.Artefact` otherwise.

    Raises:
        TypeError:
            The output types could not be determined and were not given.

    """
    # Attempt to infer output
    if out is None:
        out = output_types(fun)

    opt = get_options(opt)
    db, store = get_connection(connection)
    inputs = {}
    for k, arg in enumerate(inp):
        inputs[f"in{k}"] = _artefact(db, store, arg)

    in_types = dict([(k, val.kind) for k, val in inputs.items()])

    noutputs = len(out)
    out_type = dict([(f"out{k}", out[k]) for k in range(noutputs)])
    out_keys = list(out_type.keys())
    in_keys = list(in_types.keys())

    if name is not None:
        fun_name = name
    else:
        fun_name = f"mapping_{len(inp)}:{fun.__qualname__}"

    def __map(inpd: Mapping[str, _Data]) -> dict[str, _Data]:
        """Perform a reduction."""
        args = [inpd[key] for key in in_keys]
        out = fun(*args)
        if noutputs == 1:
            out = (out,)
        return dict(zip(out_keys, out))

    funsie = python_funsie(__map, in_types, out_type, name=fun_name, strict=strict)
    operation = make_op(db, funsie, inputs, opt)
    returnval = tuple(
        [Artefact.grab(db, operation.out[o]) for o in out_keys]  # type:ignore
    )
    if len(returnval) == 1:
        return returnval[0]
    else:
        return returnval


# --------------------------------------------------------------------------------
# Convenience functions


def morph(
    fun: Callable[[Tin1], Tout1],
    inp: Union[Tin1, Artefact[Tin1]],
    *,  # noqa:DAR101,DAR201
    out: Optional[Encoding] = None,
    name: Optional[str] = None,
    strict: bool = True,
    opt: Optional[Options] = None,
    connection: Connection = None,
) -> Artefact[Tout1]:
    """Add to workflow a one-to-one python function `y = f(x)`.

    This is syntactic sugar around `py()`. By default, the output type will
    match the input type if it can't be inferred, but it can be set to a given
    `types.Encoding` using the `out=` keyword.
    """
    db, store = get_connection(connection)
    inp2 = _artefact(db, store, inp)
    if out is None:
        try:
            typ = output_types(fun)
        except TypeError:
            typ = (inp2.kind,)

        if len(typ) > 1:
            raise TypeError(
                "Attempted to use morph but the function has more than one output.\n"
                + f"inferred return value: {typ}"
            )
        else:
            out = typ[0]

    if name is not None:
        morpher_name = name
    else:
        morpher_name = f"morph:{fun.__qualname__}"
    out_type = [out]
    return py(
        fun,
        inp2,
        out=out_type,
        name=morpher_name,
        strict=strict,
        opt=opt,
        connection=(db, store),
    )


def reduce(
    fun: Callable[..., Tout1],
    *inp: _Target,  # noqa:DAR101,DAR201
    out: Optional[Encoding] = None,
    name: Optional[str] = None,
    strict: bool = True,
    opt: Optional[Options] = None,
    connection: Connection = None,
) -> Artefact[Tout1]:
    """Add to workflow a many-to-one python function `y = f(*x)`.

    This is syntactic sugar around `py()`. By default, the output encoding is
    inferred, and if this fails, is set to match the encoding of the
    arguments if they are all the same. Output encoding can also be
    explicitly set to a given `types.Encoding` using the `out=` keyword.

    """
    inps = list(inp)
    db, store = get_connection(connection)
    inps2 = [_artefact(db, store, inp) for inp in inps]
    if out is None:
        try:
            typ = output_types(fun)
        except TypeError:
            typ = tuple(set(el.kind for el in inps2))
            if len(typ) > 1:
                raise TypeError(
                    "Inference failed for function reduce(): more than one input type was"
                    + " passed but no out= encoding.\n"
                    + "Either explicitly set return with out= or ensures all inputs "
                    + "have the same encoding.\n"
                    + f"args: {list(el.kind for el in inps2)}\n"
                    + f"inferred possible return values: {typ}"
                )

        if len(typ) > 1:
            raise TypeError(
                "Attempted to use reduce but the function has more than one output.\n"
                + f"inferred return value: {typ}"
            )
        else:
            out = typ[0]

    if name is not None:
        morpher_name = name
    else:
        morpher_name = f"reduce:{fun.__qualname__}"
    out_type = [out]
    return py(
        fun,
        *inps2,
        out=out_type,
        name=morpher_name,
        strict=strict,
        opt=opt,
        connection=(db, store),
    )
